@@ -6,7 +6,8 @@ from .mpvplayer import MpvPlayer as Player
 from .playlist import PlaylistMode, Playlist
 from .metadata_assembler import MetadataAssembler
 from .fm import FM
-from .radio import SongRadio
+from .radio import SongRadio, SongsRadio
+from .ai_radio import AIRadio, AI_RADIO_SUPPORTED
 from .lyric import LiveLyric, parse_lyric_text, Line as LyricLine, Lyric
 from .recently_played import RecentlyPlayed
 from .delegate import PlayerPositionDelegate
@@ -22,6 +23,9 @@ __all__ = (
     'FM',
     'PlaylistMode',
     'SongRadio',
+    'SongsRadio',
+    'AIRadio',
+    'AI_RADIO_SUPPORTED',
 
     'Player',
     'Playlist',
